// Fetch-on-demand conference object.
//
// $Id: conference.cc,v 1.35 2000/10/21 12:05:51 ceder Exp $
// Copyright (C) 1994  Per Cederqvist and Inge Wallin
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
// 
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
// 
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.

#include <config.h>

#include <cassert>
#include <cstdlib>

#include "conference.h"
#include "conf-cache.h"
#include "set_conf_type_q.h"
#include "change_name_q.h"
#include "set_permitted_submitters_q.h"

// ================================================================
// ================================================================
//                            Class Conference


// ================================================================ 
//                    Constructors and Destructors                  


Conference::Conference(Conf_no confno, Conf_cache * cc)
    : cno(confno),
      cache(cc)
{
}

 
Conference::Conference(const Conference &conf)
    : Fetchable(conf),
      cno(conf.cno),
      cache(conf.cache)
{
}


Conference::~Conference()
{
}


// ================================================================ 
//

const Conference & 
Conference::operator = (const Conference &c)
{
    cno = c.cno;
    cache = c.cache;
    return *this;
}


Status
Conference::prefetch() const
{
    return cache->prefetch(cno);
}

bool
Conference::prefetch_pending() const
{
    return cache->prefetch_pending(cno);
}

Status
Conference::prefetch_name() const
{
    // FIXME: the cache should be able to cache name-only information.
    return cache->prefetch(cno);
}

Kom_err
Conference::kom_errno() const
{
    return cache->get(cno).kom_errno();
}

bool
Conference::present() const
{
    return cache->present(cno);
}

bool
Conference::exists() const
{
    return kom_errno() == KOM_NO_ERROR;
}

Conf_no
Conference::conf_no() const
{
    return cno;
}

Connection *
Conference::connection() const
{	
    return cache->connection();
}

Pers_no 
Conference::creator() const
{
    assert(kom_errno() == KOM_NO_ERROR);
    return cache->get(cno).creator();
}

struct tm 
Conference::creation_time() const
{
    assert(kom_errno() == KOM_NO_ERROR);
    return cache->get(cno).creation_time();
}

Text_no 
Conference::presentation() const
{
    assert(kom_errno() == KOM_NO_ERROR);
    return cache->get(cno).presentation();
}

Conf_no 
Conference::supervisor() const
{
    assert(kom_errno() == KOM_NO_ERROR);
    return cache->get(cno).supervisor();
}

Conf_no 
Conference::permitted_submitters() const
{
    assert(kom_errno() == KOM_NO_ERROR);
    return cache->get(cno).permitted_submitters();
}

Conf_no 
Conference::super_conf() const
{
    assert(kom_errno() == KOM_NO_ERROR);
    return cache->get(cno).super_conf();
}

Conf_type 
Conference::type() const
{
    assert(kom_errno() == KOM_NO_ERROR);
    return cache->get(cno).type();
}

struct std::tm 
Conference::last_written() const
{
    assert(kom_errno() == KOM_NO_ERROR);
    return cache->get(cno).last_written();
}

Text_no 
Conference::msg_of_day() const
{
    assert(kom_errno() == KOM_NO_ERROR);
    return cache->get(cno).msg_of_day();
}

Garb_nice 
Conference::nice() const
{
    assert(kom_errno() == KOM_NO_ERROR);
    return cache->get(cno).nice();
}

LyStr 
Conference::name() const
{
    assert(kom_errno() == KOM_NO_ERROR);
    return cache->get(cno).name();
}

unsigned short 
Conference::no_of_members() const
{
    assert(kom_errno() == KOM_NO_ERROR);
    return cache->get(cno).no_of_members();
}

Local_text_no 
Conference::first_local_no() const
{
    assert(kom_errno() == KOM_NO_ERROR);
    return cache->get(cno).first_local_no();
}

unsigned long 
Conference::no_of_texts() const
{
    assert(kom_errno() == KOM_NO_ERROR);
    return cache->get(cno).no_of_texts();
}

Kom_err
Conference::set_type(Conf_type conf_type)
{
    set_conf_type_question   sct_q(cache->connection(), cno, conf_type);
    if (sct_q.receive() == st_ok) {
	if (present())
	    cache->get(cno).set_type(conf_type);
	return KOM_NO_ERROR;
    } else
	return sct_q.error();
}

Kom_err
Conference::change_name(const LyStr &new_name)
{
    change_name_question  cn_q(cache->connection(), cno, new_name);
    if (cn_q.receive() == st_ok) {
	if (present())
	    cache->get(cno).change_name(new_name);
	return KOM_NO_ERROR;
    } else
	return cn_q.error();
}

Kom_err
Conference::set_permitted_submitters(Conf_no permsub)
{
    set_permitted_submitters_question q(cache->connection(), cno, permsub);
    if (q.receive() == st_ok) 
    {
	if (present())
	    cache->get(cno).set_permitted_submitters(permsub);
	return KOM_NO_ERROR;
    } 
    else
	return q.error();
}


// ================================================================
// ================================================================
//                            Class Conferences


// ================================================================ 
//                    Constructors and Destructors                  


Conferences::Conferences(Connection *co, Async_dispatcher *async_disp)
  : conn(co),
    conf_cache(new Conf_cache(co, async_disp))
{
}
 

// Conferences::Conferences(const Conferences &conf)
//     : conn(conf.conn)
// {
//    // Warning! A refernce counting conf_cache or something similar
//    // is needed if this method is to be implemented!
//     abort();
// }


Conferences::~Conferences()
{
    if (conf_cache)
	delete conf_cache;
}


// ================================================================ 
//

Connection *
Conferences::connection() const
{
    return conn;
}

Conference
Conferences::operator [](Conf_no confno) const
{
    assert(conf_cache != NULL);
    return Conference(confno, conf_cache);
}

Conf_cache * 
Conferences::read_conf_cache() const
{
    return conf_cache;
}

