// Fetch-on-demand map from local to global text numbers.
//
// $Id: text-mapping.cc,v 1.21 2005/04/02 21:34:07 ceder Exp $
// Copyright (C) 1994  Per Cederqvist and Inge Wallin
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
// 
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
// 
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.

#include <config.h>

#include <map>
#include <algorithm>
#include <cassert>

#include "conference.h"
#include "question.h"
#include "Local2Global.h"
#include "text-mapping.h"
#include "get_map_q.h"
#include "local-text-no-set.h"

struct Text_mapping::TMI {
  public:
    TMI(Conference c, int chunk_sz);
    ~TMI();

    Conference conf;
    const int chunk_size;
    
    void add_ref();
    int sub_ref();		// Returns the new reference count.

    typedef std::map<Local_text_no, Text_no> L2g;

    L2g l2g;
    Local_text_no_set known;
    std::vector<Pending> pending;

    // These start out to cover the entire possible interval.  The
    // interval is moved as new information becomes available.
    Local_text_no first_existing;
    Local_text_no first_non_existing;

  private:
    TMI(const TMI&);		// N/A
    TMI& operator=(const TMI&); // N/A

    int refcount;
};


Text_mapping::Text_mapping()
    : tmi(NULL)
{
}

Text_mapping::Text_mapping(Conference c, int chunk_sz)
  : tmi(new TMI(c, chunk_sz))
{
    tmi->add_ref();
}

Text_mapping::Text_mapping(const Text_mapping&m)
  : tmi(m.tmi)
{
    if (tmi)
	tmi->add_ref();
}

Text_mapping::~Text_mapping()
{
    if (tmi && tmi->sub_ref() == 0)
	delete tmi;
}

Text_mapping& 
Text_mapping::operator=(const Text_mapping&m)
{
    if (this != &m)
    {
	if (tmi && tmi->sub_ref() == 0)
	    delete tmi;
	tmi = m.tmi;
	if (tmi)
	    tmi->add_ref();
    }
    return *this;
}

bool
Text_mapping::valid() const
{
    return tmi != NULL;
}

#if 0
Local_text_no
Text_mapping::end_no()
{
    // FIXME: this is sometimes very wrong.
    return mci->first_non_existing;
}

Local_text_no
Text_mapping::begin_no()
{
    // FIXME: this is sometimes very wrong.
    return mci->first_existing;
}
#endif

class Text_mapping::Pending {
  public:
    Pending(Local_text_no f, get_map_question *qp)
      : first(f), q(qp) {}
	
    Local_text_no first;
    get_map_question *q;

};

Text_no
Text_mapping::operator [](Local_text_no lno)
{
    if (tmi==NULL)		// FIXME-throw
	return 0;
    
    if (lno < tmi->first_existing || tmi->first_non_existing <= lno)
	return 0;

    poll_pending();

    {
	TMI::L2g::iterator d = tmi->l2g.find(lno);
	if (d != tmi->l2g.end())
	    return (*d).second;
    }
    
    if (tmi->known.present(lno))
	// We know about this number, but it wasn't present.  It must 
	// be a deleted text.
	return 0;

    {
	std::vector<Pending>::iterator i = get_prefetch(lno);
	(*i).q->receive();
    }

    poll_pending();

    {
	TMI::L2g::iterator d = tmi->l2g.find(lno);
	if (d != tmi->l2g.end())
	    return (*d).second;
    }

    assert(tmi->known.present(lno));
    return 0;
}

Status
Text_mapping::prefetch(Local_text_no lno)
{
    if (tmi == NULL)		// FIXME-throw
	return st_error;

    if (lno < tmi->first_existing || tmi->first_non_existing <= lno)
	return st_ok;

    poll_pending();
    if (tmi->known.present(lno))
	return st_ok;

    std::vector<Pending>::iterator i = get_prefetch(lno);
    return (*i).q->status();
}

void
Text_mapping::add(Local_text_no key, Text_no text)
{
    assert(tmi != NULL);

    if (text)
	tmi->l2g[key] = text;
    tmi->known.add(key);
}

void
Text_mapping::add(const Local2Global& l2g)
{
    assert(tmi != NULL);
    for (Local_text_no l = l2g.first(); l < l2g.first() + l2g.size(); l++)
	if (l2g[l])
	    tmi->l2g[l] = l2g[l];
    tmi->known.add(l2g.first(), l2g.first() + l2g.size());
}

void
Text_mapping::add_zero_range(Local_text_no begin,
			     Local_text_no end)
{
    assert(tmi != NULL);
    tmi->known.add(begin, end);
}

std::vector<Text_mapping::Pending>::iterator 
Text_mapping::get_prefetch(Local_text_no lno)
{
    assert(tmi != NULL);

    for (std::vector<Pending>::iterator res = tmi->pending.begin();
	 res != tmi->pending.end();
	 ++res)
    {
	if ((*res).first <= lno && lno < (*res).first + tmi->chunk_size)
	    return res;
    }
    
    tmi->pending.push_back(Pending(lno, new get_map_question(
	tmi->conf.connection(), tmi->conf.conf_no(), lno, tmi->chunk_size)));
    return tmi->pending.end() - 1;
}

struct Text_mapping::got_reply : public std::unary_function<Text_mapping::Pending &, bool> {

  got_reply(Text_mapping *t) : tm(t) {}

  bool operator () (Text_mapping::Pending &pend) 
  { 
    if (pend.q->status() != st_pending)
    {
	if (pend.q->status() == st_ok)
	    tm->add(pend.q->result());
	else
	    tm->add_zero_range(pend.first, pend.first + tm->chunk_size());
	delete pend.q;
	pend.q = NULL;
	return true;
    }
    else
	return false;
  }
private:
  Text_mapping *tm;
};

void
Text_mapping::poll_pending()
{
    std::vector<Pending>::iterator iter = std::remove_if(
	tmi->pending.begin(), tmi->pending.end(), got_reply(this));
    
    tmi->pending.erase(iter, tmi->pending.end());
}

int
Text_mapping::chunk_size() const 
{
    return tmi->chunk_size;
}

Text_mapping::TMI::TMI(Conference c, int chunk_sz)
  : conf(c), 
    chunk_size(chunk_sz),
    l2g(),
    known(),
    pending(),
    first_existing(0), 
    first_non_existing((unsigned)-1), // FIXME: ugly
    refcount(0)
{
}

Text_mapping::TMI::~TMI()
{
    // FIXME: can this leak pending[x].q?
    // FIXME: can this leak anything else?
}

void
Text_mapping::TMI::add_ref()
{
    ++refcount;
}

int
Text_mapping::TMI::sub_ref()
{
    return --refcount;
}
